package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.espec.bundle.ESpecBundle;
import be.iminds.ilabt.jfed.espec.filefetcher.ExperimentSpecificationFileManager;
import be.iminds.ilabt.jfed.espec.filefetcher.FileFetcher;
import be.iminds.ilabt.jfed.espec.model.ExperimentSpecification;
import be.iminds.ilabt.jfed.espec.util.ESpecLogListener;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.experiment.util.ExperimentRestoreInformation;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.beans.property.ReadOnlyObjectProperty;
import javafx.beans.property.ReadOnlyObjectWrapper;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.collections.ObservableMap;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.KeyPair;
import java.time.Duration;
import java.time.Instant;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * User: twalcari
 * Date: 1/30/2015
 * Time: 13:24
 */
public class Experiment {
    private static final Logger LOG = LoggerFactory.getLogger(Experiment.class);

    private static Duration DEFAULT_MAX_WAIT_UNTIL_READY = Duration.ofMinutes(10);

    /*
    Variables with basic information. These fields will always be set in the constructor
     */
    @Nonnull
    private final String name;
    @Nullable
    private final String subAuthName; //may be null

    @Nullable
    private final RequestRspecSource newRequestRspecSource;

    @Nullable
    private final RequestRspecSource existingRequestRspecSource;

    @Nullable private final ExperimentSpecification experimentSpecification;
    @Nullable private final ESpecBundle experimentSpecificationBundle;
    @Nullable private final ExperimentSpecificationFileManager experimentSpecificationFileManager;
    @Nonnull private final ESpecLogger eSpecLogger;

    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    @Nullable
    private final Instant requestedStartTime;

    /**
     * No sub-second precision allowed (FEDIBBTDEV-738). Must always be truncated to second precision!
     */
    @Nullable
    private final Instant requestedEndTime;

    @Nullable
    private Duration checkReadyInterval = null;
    @Nullable
    private Duration maxWaitUntilReady = DEFAULT_MAX_WAIT_UNTIL_READY;

    /**
     * UserSpecs that must be given access to the experiment resources
     */
    @Nullable
    private List<UserSpec> userSpecs;
    private final boolean requestSla;

    private ExperimentRestoreInformation experimentRestoreInformation;

    /*
    Variables which are subject to change troughout the lifecycle of an experiment
     */
    private final ReadOnlyObjectWrapper<Slice> slice = new ReadOnlyObjectWrapper<>();

    private final ReadOnlyObjectWrapper<ExperimentState> experimentState;

    private final ObservableList<ExperimentPart> experimentParts =
            FXCollections.observableList(new CopyOnWriteArrayList<>());
// not good enough            FXCollections.synchronizedObservableList(FXCollections.observableArrayList());

    /*
    Listeners
     */
    private final List<ExperimentChangeListener> experimentChangeListeners = new ArrayList<>();

    private final ObservableMap<String, Object> extras = FXCollections.observableHashMap();

    private final ObservableMap<DistributeSshKeypair, KeyPair> keypairs = FXCollections.observableHashMap();



    /**
     * Constructor used when starting a new experiment in a new to-be-created slice
     */
    public Experiment(@Nonnull String name, @Nullable String subAuthName, @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, boolean requestSla) {
        assert name != null;
        assert newRequestRspecSource != null;
        assert requestedEndTime != null;
        this.name = name;
        this.subAuthName = subAuthName;
        this.userSpecs = userSpecs;
        this.requestSla = requestSla;

        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = null;

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.PENDING);

        finishConstruction();
    }

    /**
     * Constructor used when starting a new experiment in a new to-be-created slice, using an ExperimentSpecification
     */
    public Experiment(@Nonnull String name, @Nullable String subAuthName,
                      @Nonnull ESpecBundle experimentSpecificationBundle,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, boolean requestSla) {
        assert name != null;
        assert requestedEndTime != null;
        this.name = name;
        this.subAuthName = subAuthName;
        this.userSpecs = userSpecs;
        this.requestSla = requestSla;

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.experimentSpecificationBundle = experimentSpecificationBundle;
        this.experimentSpecification = experimentSpecificationBundle.getExperimentSpecification();
        this.eSpecLogger = new ESpecLogger();
        this.experimentSpecificationFileManager = new ExperimentSpecificationFileManager(experimentSpecification, experimentSpecificationBundle, eSpecLogger);
        String requestRspec;
        try {
            requestRspec = new String(new FileFetcher(experimentSpecification.getRspecs().get(0).getSource(), experimentSpecificationBundle).fetchBytes(),
                    StandardCharsets.UTF_8);
        } catch (IOException e) {
            throw new RuntimeException("Failed to get Request RSpec from Experiment Specification");
        }
        this.newRequestRspecSource = new RequestRspecSource(requestRspec, ModelRspecType.BASIC);
        this.existingRequestRspecSource = null;

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.PENDING);

        finishConstruction();
    }


    /**
     * Constructor used when starting a new experiment in an existing slice
     */
    public Experiment(@Nonnull Slice slice, @Nonnull RequestRspecSource newRequestRspecSource,
                      @Nullable Instant requestedStartTime, @Nonnull Instant requestedEndTime,
                      @Nullable List<UserSpec> userSpecs, @Nullable RequestRspecSource existingRequestRspecSource, boolean requestSla) {
        this.requestSla = requestSla;
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();

        this.requestedStartTime =
                requestedStartTime != null ? requestedStartTime.truncatedTo(ChronoUnit.SECONDS) : null;
        this.requestedEndTime =
                requestedEndTime.truncatedTo(ChronoUnit.SECONDS); //No sub-second precision allowed (FEDIBBTDEV-738)

        this.newRequestRspecSource = newRequestRspecSource;
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();

        this.userSpecs = userSpecs;

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.RESTORING);
        registerSlice(slice);

        finishConstruction();
    }


    /**
     * Constructor used when restoring a slice.
     */
    public Experiment(@Nonnull Slice slice, @Nullable Instant requestedStartTime) {
        this.name = slice.getName();
        this.subAuthName = slice.getSubAuthorityName();


        this.experimentSpecification = null;
        this.experimentSpecificationBundle = null;
        this.experimentSpecificationFileManager = null;
        this.eSpecLogger = new ESpecLogger();
        this.newRequestRspecSource = null;

        RequestRspecSource existingRequestRspecSource = slice.getRequestRspec();
        if (existingRequestRspecSource == null && slice.getManifestRspec() != null && slice.getManifestRspec().getStringRspec() != null) {
            existingRequestRspecSource =
                    new RequestRspecSource(slice.getManifestRspec().getStringRspec(), ModelRspecType.FX);
        }
        this.existingRequestRspecSource = existingRequestRspecSource;

        this.requestedStartTime = requestedStartTime; //to be fixed slice.getStartDate();
        this.requestedEndTime = null;


        this.userSpecs = null;
        this.requestSla = false;
        registerSlice(slice);

        this.experimentState = new ReadOnlyObjectWrapper<>(ExperimentState.RESTORING);

        finishConstruction();
    }

    //instance initializer
    private void finishConstruction() {
        this.experimentState.addListener((observable, oldValue, newValue) -> experimentChangeListeners.forEach(ecl -> ecl.onExperimentStateChange(newValue)));

        this.experimentParts.addListener((ListChangeListener<ExperimentPart>) c -> {
            while (c.next()) {
                assert c.wasAdded() : "ExperimentParts should only be added!";
                for (ExperimentPart part : c.getAddedSubList()) {
                    experimentChangeListeners.forEach(ecl -> ecl.onExperimentPartAdded(part));
                }
            }
        });
    }

    public void registerSlice(Slice slice) {
        if (this.slice.get() != null)
            throw new IllegalStateException("Slice is already set in Experiment");
        if (slice == null)
            throw new IllegalArgumentException("Slice is null");
        if (slice.getUrn() == null)
            throw new IllegalArgumentException("Slice has no URN");
        this.slice.setValue(slice);

        this.experimentRestoreInformation = new ExperimentRestoreInformation(slice.getUrn());
        experimentRestoreInformation.bindToExperiment(this);
    }

    public void updateUserSpecs(@Nonnull List<UserSpec> userSpecs){
        this.userSpecs = userSpecs;
    }

    /**
     * Set the maximum time to wait after starting an experiment, until it becomes is ready.
     * If the experiment takes longer, the wait is ended, and the experiment is caonsidered failed.
     *
     * @param maxWaitUntilReady the maximum time to wait after starting an experiment, until it becomes is ready. This may be null, in which case there is no time limit.
     */
    public void setMaxWaitUntilReady(@Nullable Duration maxWaitUntilReady) {
        this.maxWaitUntilReady = maxWaitUntilReady;
    }

    /**
     * Set the time to wait between checking the status of an experiment.
     *
     * @param checkReadyInterval the time to wait between checking the status of an experiment. This may be null, in which case an internal default is used.
     */
    public void setCheckReadyInterval(@Nullable Duration checkReadyInterval) {
        this.checkReadyInterval = checkReadyInterval;
    }

    public void addExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (!this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.add(experimentChangeListener);
    }

    public void removeExperimentChangeListener(ExperimentChangeListener experimentChangeListener) {
        assert (this.experimentChangeListeners.contains(experimentChangeListener));

        this.experimentChangeListeners.remove(experimentChangeListener);
    }


    @Nonnull
    public String getName() {
        return name;
    }

    @Nullable
    public List<UserSpec> getUserSpecs() {
        return userSpecs;
    }

    @Nullable
    public Duration getMaxWaitUntilReady() {
        return maxWaitUntilReady;
    }

    @Nullable
    public Duration getCheckReadyInterval() {
        return checkReadyInterval;
    }

    public boolean isRequestSla() {
        return requestSla;
    }

    @Nullable
    public Instant getRequestedEndTime() {
        return requestedEndTime;
    }

    @Nullable
    public Instant getRequestedStartTime() {
        return requestedStartTime;
    }

    @Nullable
    public RequestRspecSource getNewRequestRspecSource() {
        return newRequestRspecSource;
    }

    @Nullable
    public ExperimentSpecification getExperimentSpecification() {
        return experimentSpecification;
    }

    @Nullable
    public ExperimentSpecificationFileManager getExperimentSpecificationFileManager() {
        return experimentSpecificationFileManager;
    }

    @Nonnull
    public ESpecLogger getExperimentSpecificationLogger() {
        return eSpecLogger;
    }
    public void addESpecLogListener(@Nonnull ESpecLogListener listener) {
        eSpecLogger.addESpecLogListener(listener);
    }
    public void removeESpecLogListener(@Nonnull ESpecLogListener listener) {
        eSpecLogger.removeESpecLogListener(listener);
    }

    @Nullable
    public ESpecBundle getExperimentSpecificationBundle() {
        return experimentSpecificationBundle;
    }

    @Nullable
    public Slice getSliceOrNull() {
        return slice.get();
    }

    @Nonnull
    public Optional<Slice> getSliceOpt() {
        return Optional.ofNullable(slice.get());
    }

    @Nonnull
    /**
     * This returns the slice related to this experiment, or throws an exception if there is none.
     *
     * @return the slice related to this experiment (never null)
     */
    public Slice getSlice() {
        Slice res = slice.get();
        if (res == null) {
            throw new IllegalStateException("The Experiment does not have a slice");
        }
        return res;
    }

    @Nullable
    public GeniUrn getSliceUrn() {
        return slice.get() == null ? null : slice.get().getUrn();
    }

    @Nullable
    public String getSliceUrnString() {
        return slice.get() == null ? null : slice.get().getUrnString();
    }

    public ReadOnlyObjectProperty<Slice> sliceProperty() {
        return slice.getReadOnlyProperty();
    }

    @Nonnull
    public ExperimentState getExperimentState() {
        return experimentState.get();
    }

    @Nonnull
    public ReadOnlyObjectProperty<ExperimentState> experimentStateProperty() {
        return experimentState.getReadOnlyProperty();
    }

    @Nullable
    public String getSubAuthName() {
        return subAuthName;
    }

    public boolean isReservationMade() {
        return experimentParts.stream()
                .anyMatch(experimentPart -> experimentPart.getState() == InternalState.RESERVED);
    }

    public ExperimentRestoreInformation getExperimentRestoreInformation() {
        return experimentRestoreInformation;
    }

    public void setExperimentState(@Nonnull ExperimentState experimentState) {
        LOG.debug("Setting state for experiment {} to {}", name, experimentState);
        this.experimentState.set(experimentState);

    }

    @Nullable
    public RequestRspecSource getExistingRequestRspecSource() {
        return existingRequestRspecSource;
    }

    @Nonnull
    public ObservableList<ExperimentPart> getParts() {
        return experimentParts;
    }

    public void addExperimentPart(ExperimentPart experimentPart) {
        //noinspection ObjectEquality
        assert experimentPart.getExperiment() == this;

        this.experimentParts.add(experimentPart);
    }

    @Nullable
    public SfaExperimentPart getPart(Server auth) {
        return experimentParts.stream()
                .filter(part -> part instanceof SfaExperimentPart &&
                        Objects.equals(((SfaExperimentPart) part).getConnectSfaAuthority(), auth))
                .map(SfaExperimentPart.class::cast)
                .findFirst().orElse(null);
    }

    @Nullable
    public ExperimentPart getPart(Sliver sliver) {
        return experimentParts.stream()
                .filter(part -> part instanceof SfaExperimentPart
                        && ((SfaExperimentPart) part).getSlivers().contains(sliver))
                .findFirst().orElse(null);
    }

    @Nullable
    public ExperimentPart getPart(GeniUrn sliverUrn) {
        for (ExperimentPart experimentPart : experimentParts) {
            if (experimentPart instanceof SfaExperimentPart) {
                SfaExperimentPart sfaExperimentPart = (SfaExperimentPart) experimentPart;

                for (Sliver sliver : sfaExperimentPart.getSlivers()) {
                    if (Objects.equals(sliver.getUrn(), sliverUrn))
                        return experimentPart;
                }
            }
        }
        return null;
    }


    public ObservableMap<String, Object> getExtras() {
        return extras;
    }

    public ObservableMap<DistributeSshKeypair, KeyPair> getKeypairs() {
        return keypairs;
    }
}
